@@ -43,9 +43,13 @@ using namespace android;
 
 namespace {
     // An arbitrary time value.
-    const nsecs_t arbitrary_time = 1234;
+    constexpr const std::chrono::nanoseconds arbitrary_time = std::chrono::nanoseconds(1234);
 } // anonymous namespace
 
+int const FakeEventHub::USBTouchscreenID = droidinput::BUILT_IN_KEYBOARD_ID + 2;
+int const FakeEventHub::TouchScreenMinAxisValue = 0;
+int const FakeEventHub::TouchScreenMaxAxisValue = 100;
+
 FakeEventHub::FakeEventHub()
 {
     keymap.loadGenericMaps();
@@ -221,6 +225,10 @@ size_t FakeEventHub::getEvents(int timeoutMillis, RawEvent* buffer, size_t buffe
     (void) timeoutMillis;
     {
         std::lock_guard<std::mutex> lg(guard);
+
+        if (throw_in_get_events)
+            throw std::runtime_error("FakeEventHub::getEvents() exception");
+
         for (size_t i = 0; i < bufferSize && events_available.size() > 0; ++i)
         {
             buffer[i] = events_available.front();
@@ -381,7 +389,7 @@ bool FakeEventHub::setKeyboardLayoutOverlay(int32_t deviceId,
     return true;
 }
 
-void FakeEventHub::vibrate(int32_t deviceId, nsecs_t duration)
+void FakeEventHub::vibrate(int32_t deviceId, std::chrono::nanoseconds duration)
 {
     (void)deviceId;
     (void)duration;
@@ -416,7 +424,7 @@ void FakeEventHub::flush()
 void FakeEventHub::synthesize_builtin_keyboard_added()
 {
     RawEvent event;
-    event.when = 0;
+    event.when = std::chrono::nanoseconds(0);
     event.deviceId = BuiltInKeyboardID;
     event.type = EventHubInterface::DEVICE_ADDED;
 
@@ -427,7 +435,7 @@ void FakeEventHub::synthesize_builtin_keyboard_added()
 void FakeEventHub::synthesize_builtin_cursor_added()
 {
     RawEvent event;
-    event.when = 0;
+    event.when = std::chrono::nanoseconds(0);
     event.deviceId = BuiltInCursorID;
     event.type = EventHubInterface::DEVICE_ADDED;
 
@@ -435,10 +443,38 @@ void FakeEventHub::synthesize_builtin_cursor_added()
     events_available.push_back(event);
 }
 
+void FakeEventHub::synthesize_usb_touchscreen_added()
+{
+    FakeDevice device;
+    device.classes = INPUT_DEVICE_CLASS_TOUCH | INPUT_DEVICE_CLASS_EXTERNAL;
+    device.input_properties[INPUT_PROP_DIRECT] = true;
+    device.identifier.bus = BUS_USB;
+
+    RawAbsoluteAxisInfo axis_info;
+    axis_info.valid = true;
+    axis_info.minValue = TouchScreenMinAxisValue;
+    axis_info.maxValue = TouchScreenMaxAxisValue;
+    axis_info.flat = 0;
+    axis_info.fuzz = 0;
+    axis_info.resolution = 1;
+    device.absoluteAxes.add(ABS_X, axis_info);
+    device.absoluteAxes.add(ABS_Y, axis_info);
+
+    device_from_id.insert(std::pair<int32_t, FakeDevice>(USBTouchscreenID, device));
+    
+    RawEvent event;
+    event.when = std::chrono::nanoseconds(0);
+    event.deviceId = USBTouchscreenID;
+    event.type = EventHubInterface::DEVICE_ADDED;
+    
+    std::lock_guard<std::mutex> lg(guard);
+    events_available.push_back(event);
+}
+
 void FakeEventHub::synthesize_device_scan_complete()
 {
     RawEvent event;
-    event.when = 0;
+    event.when = std::chrono::nanoseconds(0);
     event.type = EventHubInterface::FINISHED_DEVICE_SCAN;
 
     std::lock_guard<std::mutex> lg(guard);
@@ -448,7 +484,7 @@ void FakeEventHub::synthesize_device_scan_complete()
 void FakeEventHub::synthesize_event(const mis::KeyParameters &parameters)
 {
     RawEvent event;
-    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch()).count();
+    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch());
     event.type = EV_KEY;
     event.code = parameters.scancode;
 
@@ -469,7 +505,7 @@ void FakeEventHub::synthesize_event(const mis::KeyParameters &parameters)
 void FakeEventHub::synthesize_event(const mis::ButtonParameters &parameters)
 {
     RawEvent event;
-    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch()).count();
+    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch());
     event.type = EV_KEY;
     event.code = parameters.button;
 
@@ -495,7 +531,7 @@ void FakeEventHub::synthesize_event(const mis::ButtonParameters &parameters)
 void FakeEventHub::synthesize_event(const mis::MotionParameters &parameters)
 {
     RawEvent event;
-    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch()).count();
+    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch());
     event.type = EV_REL;
     if (parameters.device_id)
         event.deviceId = parameters.device_id;
@@ -517,7 +553,40 @@ void FakeEventHub::synthesize_event(const mis::MotionParameters &parameters)
     events_available.push_back(event);
 }
 
-void FakeEventHub::synthesize_event(nsecs_t when, int32_t device_id, int32_t type, int32_t code, int32_t value)
+void FakeEventHub::synthesize_event(const mis::TouchParameters &parameters)
+{
+    RawEvent event;
+    event.when = std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch());
+    event.type = EV_ABS;
+    if (parameters.device_id)
+        event.deviceId = parameters.device_id;
+    else
+        event.deviceId = USBTouchscreenID;
+
+    std::lock_guard<std::mutex> lg(guard);
+    event.code = ABS_X;
+    event.value = parameters.abs_x;
+    events_available.push_back(event);
+
+    event.code = ABS_Y;
+    event.value = parameters.abs_y;
+    events_available.push_back(event);
+
+    if (parameters.action != mis::TouchParameters::Action::Move)
+    {
+        event.type = EV_KEY;
+        event.code = BTN_TOUCH;
+        event.value = (parameters.action == mis::TouchParameters::Action::Tap);
+        events_available.push_back(event);
+    }
+
+    // Cursor motion events require a sync as per droidinput::CursorInputMapper::process
+    event.type = EV_SYN;
+    event.code = SYN_REPORT;
+    events_available.push_back(event);
+}
+
+void FakeEventHub::synthesize_event(std::chrono::nanoseconds when, int32_t device_id, int32_t type, int32_t code, int32_t value)
 {
     RawEvent event;
     event.when = when;
@@ -686,5 +755,12 @@ const FakeEventHub::FakeDevice* FakeEventHub::getDevice(int32_t deviceId) const
 
 size_t FakeEventHub::eventsQueueSize() const
 {
+    std::lock_guard<std::mutex> lg(guard);
     return events_available.size();
 }
+
+void FakeEventHub::throw_exception_in_next_get_events()
+{
+    std::lock_guard<std::mutex> lg(guard);
+    throw_in_get_events = true;
+}
